/*
 * Copyright (c) 2013, 
 * Sangoma Technologies James Zhang (jzhang@sangoma.com)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of the original author; nor the names of any contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 */

#include "analog_tap.h"

#ifdef __WINDOWS__
typedef struct ftdm_buffer {
	unsigned char *data;
	unsigned char *head;
	ftdm_size_t used;
	ftdm_size_t actually_used;
	ftdm_size_t datalen;
	
	ftdm_size_t max_len;
	ftdm_size_t blocksize;
	unsigned id;
	int loops;
} ftdm_buffer_t;
#endif

/********************************************************************************/
/*                                                                              */
/*                           function declaration                               */
/*                                                                              */
/********************************************************************************/
#ifndef __WINDOWS__
void on_sigint(int signal);
void on_sigusr(int signal);
#endif

int sample_app_main_menu(void);
int sample_app_print_status(span_data_t * span_data);
span_data_t *sample_app_span_menu(void);

/* Set SDK_DEBUGGING since this is a sample application,
 * This should be turned off in Production */
#define SDK_DEBUGGING 1

/********************************************************************************/
/*                                                                              */
/*                              global parameters                               */
/*                                                                              */
/********************************************************************************/
int g_running = 1;
int g_num_threads = 0;

span_data_t SPANS[FTDM_MAX_SPANS_INTERFACE + 1] = { {0} };
sangoma_freetdm_init_t sample_app_ftdm_init;

/********************************************************************************/
/*                                                                              */
/*                              implementation                                  */
/*                                                                              */
/********************************************************************************/

#ifndef __WINDOWS__
void on_sigusr(int signal)
{
	switch (signal) {
	case SIGUSR1:
		fprintf(stdout, "User interrupted... \n");
		break;

	case SIGUSR2:
		fprintf(stderr, "User interrupted... \n");
		break;
	}

	fprintf(stdout, "SANGOMA ANALOG TAPPING MODULE Application is Getting closed\n");
	ftdm_global_destroy();
	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "                           GOOD BYE                            \n");
	fprintf(stdout, "=========================================================\n");
	getchar();
	fprintf(stdout, "Press enter... \n");
	exit(EXIT_FAILURE);
}

void on_sigint(int signal)
{
	g_running = 0;

	fprintf(stdout, "SIGINT recieved ...\n");
	fprintf(stdout, "SANGOMA ANALOG TAPPING MODULE Application is Getting closed\n");
	ftdm_global_destroy();
	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "                      GOOD BYE                            \n");
	fprintf(stdout, "=========================================================\n");
	fprintf(stdout, "Press Enter key to exit... \n");
	getchar();
	exit(EXIT_FAILURE);
}
#endif

/****************************************
 FreeTDM Event Callbacks
 ****************************************/

int sample_app_hwprobe_callback(void *sangoma_port_obj, int port_id, void *user_obj)
{
	int card_type = sangoma_config_port_hwinfo_get_model_id(sangoma_port_obj);
	int err;

	/* Filter out all other cards just look for W400 */
	if (card_type != A200_ADPTR_ANALOG && card_type != A400_ADPTR_ANALOG) {
		return 0;
	}

	SPANS[port_id].init = 1;
	SPANS[port_id].sangoma_port_obj = sangoma_port_obj;

	err = sangoma_config_port_stop(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to save sangoma port config\n");
	}

	err = sangoma_config_port_save(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to save sangoma port config\n");
		return -1;
	}

	ftdm_sleep(5000);

	err = sangoma_config_port_start(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to start sangoma port config\n");
		return -1;
	}

	/* Create spans in FreeTDM */
	err = sangoma_freetdm_create_span(sangoma_port_obj, &SPANS[port_id].ftdm_span, SNG_ANALOG_TAPPING_MODE);
	if (err) {
		fprintf(stderr, "Error: Failed to create freetdm span %i\n", port_id);
		return -1;
	}

	if (sangoma_freetdm_span_config_and_start(sangoma_port_obj, SPANS[port_id].ftdm_span, NULL, SNG_ANALOG_TAPPING_MODE)) {
		fprintf(stderr, "Error: Failed to load and configure freetdm span %i\n", port_id);
		return -1;
	}

	return 0;
}

/*
 * \brief Retrieves tone generation output to be sent
 * \param ts Teletone generator
 * \param map Tone map
 * \return -1 on error, 0 on success
 */
static int teletone_handler(teletone_generation_session_t *ts, teletone_tone_map_t *map)
{
	ftdm_buffer_t *dt_buffer = ts->user_data;
	int wrote;

	if (!dt_buffer) {
		return -1;
	}
	wrote = teletone_mux_tones(ts, map);
	ftdm_buffer_write(dt_buffer, ts->buffer, wrote * 2);
	return 0;
}

int analog_main_menu(void)
{
	char choice[6];
	span_data_t *span_data;

	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "		WELCOME TO SANGOMA ANALOG TAPPING MODULE             \n");
	fprintf(stdout, "=========================================================\n");

	span_data = sample_app_span_menu();

	fprintf(stdout, "Analog tapping will record calls automatically when OFFHOOK/ONHOOK events sniffered.\n");
	fprintf(stdout, "Please make calls to test. \n");
	fprintf(stdout, "Recorded file name format is (Callerid)-in-s(span id)c(chan id)-(datetime). \n");
	fprintf(stdout, "Press 1 to show span status. \n");

	if (NULL != fgets(choice, 6, stdin)) {
		if (2 < strlen(choice)) {
			fprintf(stderr, "Invalid Input Choice must must be [1-4]...\n");
			return 0;
		}

		choice[1] = '\0';

		/* Provides option to user to perform various operations like:
		 * 1) Print Span related information
		 * 2) Exit from application
		 */
		switch (choice[0]) {
		case '1':
			while (sample_app_print_status(span_data)) {
				continue;
			}
			break;

		case '2':
			fprintf(stdout, "Analog Tapping Sample Application is getting closed...\n");
			fprintf(stdout, "\n\n\n=========================================================\n");
			fprintf(stdout, "                      GOOD BYE                            \n");
			fprintf(stdout, "=========================================================\n");
			return -1;

		default:
			fprintf(stderr, "Invalid Option. Please Choose a correct option value...\n");
			break;
		}
	}

	return 0;
}

span_data_t *sample_app_span_menu(void)
{
	int idx = 0;
	char choice[128];

	while (1) {
		printf(" Span Menu \n");
		for (idx = 1; idx <= FTDM_MAX_SPANS_INTERFACE; idx++) {

			if (!SPANS[idx].ftdm_span) {
				continue;
			}

			printf("%d. %s\n", idx, ftdm_span_get_name(SPANS[idx].ftdm_span));
		}

		printf("Select Span ID or Ctrl-C for exit\n");
		printf("\n");

		if (NULL != fgets(choice, 20, stdin)) {
			if (19 < strlen(choice)) {
				fprintf(stderr, "Invalid Input...\n");
				continue;
			} else {
				break;
			}
		}
	}

	choice[strlen(choice) - 1] = '\0';
	idx = atoi(choice);

	if (idx == 0 || idx > FTDM_MAX_SPANS_INTERFACE) {
		printf("Invalid ID\n");
		return sample_app_span_menu();
	}

	if (!SPANS[idx].ftdm_span) {
		return sample_app_span_menu();
	}

	printf("Got Span %s\n", ftdm_span_get_name(SPANS[idx].ftdm_span));
	printf("\n");

	return &SPANS[idx];
}

int sample_app_print_status(span_data_t * span_data)
{
	char choice[6];
	int ret = 1;

	fprintf(stdout, "\n\n\nPlease select any one of the below options :\n");

	fprintf(stdout, "1. Physical Span list\n2. Physical Span Status\n3. Go Back To Previous Menu\n");
	fprintf(stdout, "Enter your choice[1-3]: ");

	if (NULL != fgets(choice, 6, stdin)) {
		if (2 < strlen(choice)) {
			fprintf(stderr, "Invalid Input Choice must must be [1-3]...\n");
			return -1;
		}

		choice[1] = '\0';

		switch (choice[0]) {
		case '1':
			{
				ret = sangoma_freetdm_channel_list(span_data->ftdm_span, 1);
				ftdm_sleep(100);
			}
			break;

		case '2':
			{
				ret = sangoma_freetdm_channel_dump(span_data->ftdm_span, 1);
				ftdm_sleep(100);
			}
			break;

		case '3':
			return 0;
			break;

		default:
			fprintf(stderr, "Invalid Option. Please Choose a correct option value...\n");
			break;
		}
	}

	return ret;
}

/**********************************************************************************
 * Fun   : start_analog_tapping()
 *
 * Desc  : Starts tapping the call on the respective span and channel on which
 *         call is established
 *
 * Ret   : NULL
**********************************************************************************/
static void *start_analog_tapping(ftdm_thread_t * me, void *obj)
{
	uint8_t frame[1024];
	ftdm_sigmsg_t *sigmsg = (ftdm_sigmsg_t *) obj;
	ftdm_channel_t *chan = NULL;
	ftdm_span_t *span = NULL;
	ftdm_size_t len;
	ftdm_caller_data_t *caller_data = NULL;
	uint32_t span_id, chan_id;
	char path[255];
	char out_digits[25];
	uint8_t dir;

	char time_str[128];

	memset(time_str, 0, sizeof(time_str));
	sangoma_get_timestamp_str(time_str);

	if (!sigmsg) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid signalling message\n");
		goto done;
	}

	span_id = sigmsg->span_id;
	chan_id = sigmsg->chan_id;
	memcpy(out_digits, sigmsg->ev_data.collected.digits, sizeof(out_digits));

	if (strlen(out_digits)) {
		dir = DIR_OUT;
		ftdm_channel_call_answer(sigmsg->channel);
	} else {
		dir = DIR_IN;
	}

	if (!span_id) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid Span ID\n");
		goto done;
	}

	ftdm_span_find(span_id, &span);

	if (!chan_id) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid channel ID\n");
		goto done;
	}

	if (!span) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid span found\n");
		goto done;
	}
	chan = ftdm_span_get_channel(span, chan_id);

	if (!chan) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid channel found\n");
		goto done;
	}

	ftdm_log(FTDM_LOG_DEBUG, "Start tapping on span %d and channel %d\n", span_id, chan_id);
	caller_data = ftdm_channel_get_caller_data(chan);

	if (dir == DIR_IN) {
		snprintf(path, sizeof(path), "%s-in-s%dc%d-%s", caller_data->ani.digits, span_id, chan_id, time_str);
	} else if (dir == DIR_OUT) {
		snprintf(path, sizeof(path), "%s-out-s%dc%d-%s", out_digits, span_id, chan_id, time_str);
	}

	ftdm_channel_command(chan, FTDM_COMMAND_TRACE_OUTPUT, path);

	/* Start reading the media, Due to calling ftdm_channel_read API, FreeTDM internally will write
	 * receive media to trace file(path as mentioned in above command) */
	while (ftdm_channel_call_check_answered(chan)) {
		ftdm_wait_flag_t flags = FTDM_READ;

		if (ftdm_channel_wait(chan, &flags, -1) == FTDM_FAIL) {
			break;
		}
		len = sizeof(frame);
		if (flags & FTDM_READ) {
			if (ftdm_channel_read(chan, frame, &len) == FTDM_SUCCESS) {
				/* Passive Tapping application should not write to channel, only read */
			} else {
				ftdm_log(FTDM_LOG_DEBUG, "No Data to read\n");
				break;
			}
		}
	}

  done:

	ftdm_free(sigmsg);
	return NULL;
}

static void *app_freetdm_event_queue_handler(ftdm_thread_t * me, void *obj)
{
	ftdm_sigmsg_t *sigmsg = NULL;

	teletone_generation_session_t ts;
	ftdm_buffer_t *dt_buffer = NULL;
	ftdm_channel_t *ftdmchan = NULL;
	ftdm_size_t len, rlen;
	uint8_t frame[1024];
	const char cmd[128];
	ftdm_wait_flag_t wflags = FTDM_WRITE;
	ftdm_status_t write_res;

	ts.buffer = NULL;

	if (ftdm_buffer_create(&dt_buffer, 1024, 50000, 0) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "MEM ERROR: Failed to allocate memory to buffer\n");
		return NULL;
	}

	/* initializing teletone session */
	teletone_init_session(&ts, 0, teletone_handler, dt_buffer);
	ts.rate = 8000;

	ts.debug = 1;
	ts.debug_stream = stdout;

	while (g_running) {
		sigmsg = ftdm_queue_dequeue(sample_app_ftdm_init.event_msg_queue);
		if (sigmsg) {
			ftdm_sigmsg_t *sigmsg_cpy = NULL;
			/*
			   printf("Span = %d, chan=%d, event_id=%d, call_priv*=%p \n", 
			   sigmsg->span_id, sigmsg->chan_id, 
			   sigmsg->event_id, sigmsg->call_priv
			   );
			 */


			switch (sigmsg->event_id) {
			case FTDM_SIGEVENT_START:
				{
					sigmsg_cpy = ftdm_malloc(sizeof(ftdm_sigmsg_t));
					memcpy(sigmsg_cpy, sigmsg, sizeof(ftdm_sigmsg_t));
					ftdm_log(FTDM_LOG_INFO, "SIG_START event received. Starts io data operation for [s%dc%d]. \n",
							 sigmsg->span_id, sigmsg->chan_id);

					//ftdm_channel_call_answer(sigmsg->channel);
					if (ftdm_thread_create_detached(start_analog_tapping, sigmsg_cpy) != FTDM_SUCCESS) {
						ftdm_log(FTDM_LOG_ERROR, "Not able to launch IO thread for [s%dc%d]. \n",
								 sigmsg->span_id, sigmsg->chan_id);
						break;
					}
					break;
				}
			case FTDM_SIGEVENT_UP:
				{
					ftdm_buffer_zero(dt_buffer);
					ftdmchan = sigmsg->channel;

					sigmsg_cpy = ftdm_malloc(sizeof(ftdm_sigmsg_t));
					memcpy(sigmsg_cpy, sigmsg, sizeof(ftdm_sigmsg_t));
					ftdm_log(FTDM_LOG_INFO, "SIG_UP event received. Starts io data operation for [s%dc%d]. \n",
							 sigmsg->span_id, sigmsg->chan_id);

					//ftdm_channel_call_answer(sigmsg->channel);
					if (ftdm_thread_create_detached(start_analog_tapping, sigmsg_cpy) != FTDM_SUCCESS) {
						ftdm_log(FTDM_LOG_ERROR, "Not able to launch IO thread for [s%dc%d]. \n",
								 sigmsg->span_id, sigmsg->chan_id);
						break;
					}
					ftdm_channel_command(ftdmchan, FTDM_COMMAND_OFFHOOK, NULL);
					ftdm_sleep(5000);
					ftdm_channel_command(ftdmchan, FTDM_COMMAND_SEND_DTMF, "1");
					ftdm_sleep(2000);
					ftdm_channel_command(ftdmchan, FTDM_COMMAND_SEND_DTMF, "2");
					ftdm_sleep(2000);
					ftdm_channel_command(ftdmchan, FTDM_COMMAND_SEND_DTMF, "3");
					ftdm_sleep(2000);
					ftdm_channel_command(ftdmchan, FTDM_COMMAND_SEND_DTMF, "4");
					ftdm_sleep(2000);
					ftdm_channel_command(ftdmchan, FTDM_COMMAND_SEND_DTMF, "5");

					ftdm_sleep(5000);
					strncpy(cmd, "%(2000,0,500)", 127);
					/* Generate tone and send to peer in order to recieve caller ID */
					teletone_run(&ts, cmd);

					rlen = ftdm_buffer_read(dt_buffer, frame, 160*2);
					if (!rlen) {
						ftdm_log(FTDM_LOG_DEBUG, "Reached the end of teletone buffer. \n");
						break;
					}
					do {
						if (ftdm_channel_get_codec(ftdmchan) != FTDM_CODEC_SLIN) {
							fio_codec_t codec_func = NULL;

							if (ftdm_channel_get_native_codec(ftdmchan) == FTDM_CODEC_ULAW) {
								codec_func = fio_slin2ulaw;
							} else if (ftdm_channel_get_native_codec(ftdmchan) == FTDM_CODEC_ALAW) {
								codec_func = fio_slin2alaw;
							}    

							if (codec_func) {
								codec_func(frame, sizeof(frame), &rlen);
							} else {
								ftdm_log(FTDM_LOG_ERROR, "Invalid native codec!\n");
								goto done;
							}    
						}    
keepwait:
						wflags = FTDM_WRITE;
						if (FTDM_SUCCESS != ftdm_channel_wait (ftdmchan, &wflags, -1) ) {
							ftdm_log(FTDM_LOG_CRIT, "WAIT FAILED, flags = %d\n", wflags);
							goto keepwait;
						} else if (!(wflags & FTDM_WRITE)) {
							ftdm_log(FTDM_LOG_CRIT, "Flags said device is not ready. wflags = %d\n", wflags);
							goto keepwait;
						}

						write_res = ftdm_channel_write(ftdmchan, frame, sizeof(frame), &rlen);
						if ( write_res != FTDM_SUCCESS) {
							ftdm_log(FTDM_LOG_CRIT, "Error in writing teletone frames. Err = %d \n", write_res);
						} else {
							ftdm_log(FTDM_LOG_DEBUG, "Successfully wrote teletone frames. \n");
						}
					} while (rlen = ftdm_buffer_read(dt_buffer, frame, sizeof(frame)));

					ftdm_channel_command(ftdmchan, FTDM_COMMAND_ONHOOK, NULL);



					break;
				}
			case FTDM_SIGEVENT_RELEASED:
				ftdm_log(FTDM_LOG_INFO, "SIG_RELEASED event received. Stops io data operation for [s%dc%d]. \n",
						 sigmsg->span_id, sigmsg->chan_id);
				break;
			default:
				break;
			}

			/* sample app owns the event buffer. It must be deallocated here
			   otherwise memory leak */
			ftdm_safe_free(sigmsg);
		}

done:
		/* TODO: implement wait on queue */
		ftdm_sleep(100);
	}


	if (ts.buffer) {
		teletone_destroy_session(&ts);
	}

	if (dt_buffer) {
		ftdm_buffer_destroy(&dt_buffer);
	}

	return NULL;
}

int main(void)
{
	/* local variables initialization */
	int max_wait = 20;
	int err;
	int detected = 0;
	int configured = 0;
	sangoma_hw_config_reg_t reg;

#ifndef __WINDOWS__
	/* Signal handler in order to handle different signal */
	signal(SIGINT, on_sigint);
	signal(SIGUSR1, on_sigusr);
	signal(SIGUSR2, on_sigusr);
#endif

	/* Initialize global FreeTDM init structure */
	memset(&sample_app_ftdm_init, 0, sizeof(sample_app_ftdm_init));

	/* Set basic debuggin parameters */
#ifdef SDK_DEBUGGING
	sample_app_ftdm_init.logging_level = FTDM_LOG_LEVEL_DEBUG;
	sample_app_ftdm_init.crash_policy_level = FTDM_CRASH_ON_ASSERT;
#else
	sample_app_ftdm_init.logging_level = FTDM_LOG_LEVEL_INFO;
	sample_app_ftdm_init.crash_policy_level = FTDM_CRASH_NEVER;
#endif

	/* Initilize FreeTDM Framework 
	   Global Varialbes and FreeTDM Logger */
	err = sangoma_freetdm_init(&sample_app_ftdm_init);
	if (err) {
		return err;
	}

	/* Initialize Sangoma Config SDK */
	reg.user_obj = NULL;
	reg.hwprobe_callback = sample_app_hwprobe_callback;
	reg.hwprobe_callback_post = NULL;
	err = sangoma_config_reg(&reg);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Sangoma Config SDK Initialization Error\n");
		goto exit_sample_app;
	}

	/* Create a global FreeTDM event handling queue.
	 * FreeTDM event callback function is not allowed to exectue FreeTDM API commands.
	 * FreeTDM event callback fucntion will queue events, and the app event handler thread
	 * will consume and handle FreeTDM events.  
	 * The app event handler is allowed to run FreeTDM api commands directly */
	if (ftdm_thread_create_detached(app_freetdm_event_queue_handler, NULL) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to create message queue handler thread. \n");
		goto exit_sample_app;
	}

	/* Detect Sangoma Ports 
	   For each detected port run the sample_app_hwprobe_callback
	 */
	err = sangoma_config_hwprobe(&detected, &configured, SNG_ANALOG_TAPPING_MODE);

	ftdm_log(FTDM_LOG_INFO, "Sangoma SDK Detected %d Configured %d\n", detected, configured);

	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error Failed to detect sangoma hardware \n");

		/* We cannot exit here, as the thread is launched */
		goto exit_sample_app;
	}

	/* At this point all sangoma hw ports have been configured and started.
	   And FreeTDM Spans have been creaed and started on each hw port */
	ftdm_sleep(1000);

	/* User drive main menu. 
	 * Test features for the Sangoma TDM SDK 
	 */
	while (analog_main_menu() == 0) {
		continue;
	}

  exit_sample_app:

	/* Stop all threads. 
	 * TODO: use join instead */
	g_running = 0;
	ftdm_sleep(1000);

	while (g_num_threads && (max_wait-- > 0)) {
		fprintf(stderr, "Waiting for all threads to finish (%d)\n", g_num_threads);
		ftdm_sleep(100);
	}

	/* Stop FreeTDM Framework and all Spans */
	ftdm_global_destroy();

	/* TODO: Stop all Sangoma hw ports */
	ftdm_sleep(1000);
	return 0;
}
